#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "php_wsq.h"

#if HAVE_WSQ



ZEND_DECLARE_MODULE_GLOBALS(wsq)

static zend_function_entry wsq_functions[] = {
    PHP_FE(wsq_convert_image, NULL )
    PHP_FE(wsq_read_image_from_file, NULL )
    PHP_FE(wsq_save_image_to_file, NULL )
    PHP_FE(wsq_write_wsq_bitrate, NULL )
    PHP_FE(wsq_read_wsq_bitrate, NULL )
    PHP_FE(wsq_write_wsq_ppi, NULL )
    PHP_FE(wsq_read_wsq_ppi, NULL )
    PHP_FE(wsq_write_wsq_comment, NULL )
    PHP_FE(wsq_read_wsq_comment, NULL )
    PHP_FE(wsq_write_tiff_compression, NULL )
    PHP_FE(wsq_write_tiff_predictor, NULL )
    PHP_FE(wsq_generate_serial_number, NULL )
    PHP_FE(wsq_unlock_wsq_library, NULL )
    PHP_FE(wsq_decode_stream, NULL )
    PHP_FE(wsq_encode_stream, NULL )

    {NULL, NULL, NULL}
};

zend_module_entry wsq_module_entry = {
    STANDARD_MODULE_HEADER,
    PHP_WSQ_EXTNAME,
    wsq_functions,
    PHP_MINIT(wsq),
    PHP_MSHUTDOWN(wsq),
    PHP_RINIT(wsq),
    NULL,
    NULL,
    PHP_WSQ_VERSION,
    STANDARD_MODULE_PROPERTIES
};

#ifdef COMPILE_DL_WSQ
ZEND_GET_MODULE(wsq)
#endif




// Here we define php.ini entry we expect to be
#ifdef __x86_64__
PHP_INI_BEGIN()
    // Default value is set here 
    PHP_INI_ENTRY("wsq.library_location", "/usr/lib/libWSQ_library64.so", PHP_INI_ALL, NULL)
PHP_INI_END()
#else
PHP_INI_BEGIN()
    // Default value is set here 
    PHP_INI_ENTRY("wsq.library_location", "/usr/lib/libWSQ_library.so", PHP_INI_ALL, NULL)
PHP_INI_END()
#endif



static void php_wsq_init_globals(zend_wsq_globals *wsq_globals)
{
}


 
// Declarations of WSQ library functions. Not all of them are used in extension



typedef int (*_ReadImageFromFile)(const char*, int*, int*, unsigned char**);
_ReadImageFromFile ReadImageFromFile = 0;

typedef int (*_SaveImageToFile)(const char*, int, int, int, unsigned char*);
_SaveImageToFile SaveImageToFile = 0;

typedef void (*_WriteWSQ_bitrate)(double);
_WriteWSQ_bitrate WriteWSQ_bitrate = 0;

typedef double (*_ReadWSQ_bitrate)();
_ReadWSQ_bitrate ReadWSQ_bitrate = 0;

typedef void (*_WriteWSQ_ppi)(int);
_WriteWSQ_ppi WriteWSQ_ppi = 0;

typedef int (*_ReadWSQ_ppi)();
_ReadWSQ_ppi ReadWSQ_ppi = 0;

typedef void (*_WriteWSQ_comment)(char*);
_WriteWSQ_comment WriteWSQ_comment = 0;

typedef char* (*_ReadWSQ_comment)();
_ReadWSQ_comment ReadWSQ_comment = 0;

typedef void (*_WriteTIFFcompression)(int);
 _WriteTIFFcompression WriteTIFFcompression = 0;

typedef void (*_WriteTIFFpredictor)(int);
_WriteTIFFpredictor WriteTIFFpredictor = 0;

typedef char* (*_GenerateSerialNumber)();
_GenerateSerialNumber GenerateSerialNumber = 0;

typedef int (*_UnlockWSQLibrary)(char* authorizationcode);
_UnlockWSQLibrary UnlockWSQLibrary = 0;

typedef int (*_WSQ_decode_stream)(unsigned char* in_data_stream, const int in_ds_lenght, 
	    unsigned char** out_data_stream, int* width, int* height, int* ppi, char** comment_text);
_WSQ_decode_stream WSQ_decode_stream = 0;

typedef int (*_WSQ_encode_stream)(unsigned char* in_data_stream, const int width, const int height,
	    const double bitrate, const int ppi, char* comment_text, 
	    unsigned char** out_data_stream, int* output_stream_lenght);
_WSQ_encode_stream WSQ_encode_stream = 0;



// Function which is invoked on startup of PHP
// It will open WSQ library and assign values to declared functions

PHP_RINIT_FUNCTION(wsq)
{
    WSQ_G(dso_handle) = dlopen(INI_STR("wsq.library_location"), RTLD_LAZY); // open shared library
    char* error = dlerror();
    if(error) {
	php_printf("%s\n", error);
	return FAILURE;
    }
    if (!ReadImageFromFile) // If not assigned - assign function from .so
    {
        ReadImageFromFile = (_ReadImageFromFile)dlsym(WSQ_G(dso_handle), "ReadImageFromFile");
        error = dlerror();
	if(error) {
	    php_printf("%s\n", error);
	    return FAILURE;
	}
    }
    
    if (!SaveImageToFile) // If not assigned - assign function from .so
    {
        SaveImageToFile = (_SaveImageToFile)dlsym(WSQ_G(dso_handle), "SaveImageToFile");
        error = dlerror();
	if(error) {
	    php_printf("%s\n", error);
	    return FAILURE;
	}
    }



    if (!WriteWSQ_bitrate) // If not assigned - assign function from .so
    {
        WriteWSQ_bitrate = (_WriteWSQ_bitrate)dlsym(WSQ_G(dso_handle), "WriteWSQ_bitrate");
        error = dlerror();
	if(error) {
	    php_printf("%s\n", error);
	    return FAILURE;
	}
    }


    if (!ReadWSQ_bitrate) // If not assigned - assign function from .so
    {
        ReadWSQ_bitrate = (_ReadWSQ_bitrate)dlsym(WSQ_G(dso_handle), "ReadWSQ_bitrate");
        error = dlerror();
	if(error) {
	    php_printf("%s\n", error);
	    return FAILURE;
	}
    }

    if (!WriteWSQ_ppi) // If not assigned - assign function from .so
    {
        WriteWSQ_ppi = (_WriteWSQ_ppi)dlsym(WSQ_G(dso_handle), "WriteWSQ_ppi");
        error = dlerror();
	if(error) {
	    php_printf("%s\n", error);
	    return FAILURE;
	}
    }

    if (!ReadWSQ_ppi) // If not assigned - assign function from .so
    {
        ReadWSQ_ppi = (_ReadWSQ_ppi)dlsym(WSQ_G(dso_handle), "ReadWSQ_ppi");
        error = dlerror();
	if(error) {
	    php_printf("%s\n", error);
	    return FAILURE;
	}
    }

    if (!WriteWSQ_comment) // If not assigned - assign function from .so
    {
        WriteWSQ_comment = (_WriteWSQ_comment)dlsym(WSQ_G(dso_handle), "WriteWSQ_comment");
        error = dlerror();
	if(error) {
	    php_printf("%s\n", error);
	    return FAILURE;
	}
    }

    if (!ReadWSQ_comment) // If not assigned - assign function from .so
    {
        ReadWSQ_comment = (_ReadWSQ_comment)dlsym(WSQ_G(dso_handle), "ReadWSQ_comment");
        error = dlerror();
	if(error) {
	    php_printf("%s\n", error);
	    return FAILURE;
	}
    }

    if (!WriteTIFFcompression) // If not assigned - assign function from .so
    {
        WriteTIFFcompression = (_WriteTIFFcompression)dlsym(WSQ_G(dso_handle), "WriteTIFFcompression");
        error = dlerror();
	if(error) {
	    php_printf("%s\n", error);
	    return FAILURE;
	}
    }

    if (!WriteTIFFpredictor) // If not assigned - assign function from .so
    {
        WriteTIFFpredictor = (_WriteTIFFpredictor)dlsym(WSQ_G(dso_handle), "WriteTIFFpredictor");
        error = dlerror();
	if(error) {
	    php_printf("%s\n", error);
	    return FAILURE;
	}
    }

    if (!GenerateSerialNumber) // If not assigned - assign function from .so
    {
        GenerateSerialNumber = (_GenerateSerialNumber)dlsym(WSQ_G(dso_handle), "GenerateSerialNumber");
        error = dlerror();
	if(error) {
	    php_printf("%s\n", error);
	    return FAILURE;
	}
    }

    if (!UnlockWSQLibrary) // If not assigned - assign function from .so
    {
        UnlockWSQLibrary = (_UnlockWSQLibrary)dlsym(WSQ_G(dso_handle), "UnlockWSQLibrary");
        error = dlerror();
	if(error) {
	    php_printf("%s\n", error);
	    return FAILURE;
	}
    }




    if (!WSQ_decode_stream) // If not assigned - assign function from .so
    {
        WSQ_decode_stream = (_WSQ_decode_stream)dlsym(WSQ_G(dso_handle), "WSQ_decode_stream");
        error = dlerror(); 
	if(error) {
	    php_printf("%s\n", error);
	    return FAILURE;
	}
    }

    if (!WSQ_encode_stream) // If not assigned - assign function from .so
    {
        WSQ_encode_stream = (_WSQ_encode_stream)dlsym(WSQ_G(dso_handle), "WSQ_encode_stream");
        error = dlerror(); 
	if(error) {
	    php_printf("%s\n", error);
	    return FAILURE;
	}
    }

    return SUCCESS;
}


// Init on page load

PHP_MINIT_FUNCTION(wsq)
{
    ZEND_INIT_MODULE_GLOBALS(wsq, php_wsq_init_globals, NULL);

    REGISTER_INI_ENTRIES();

    return SUCCESS;
}


// PHP shutdown function

PHP_MSHUTDOWN_FUNCTION(wsq)
{
    UNREGISTER_INI_ENTRIES();

    if ( WSQ_G(dso_handle) ) {
	dlclose(WSQ_G(dso_handle)); // close the shared library
	return SUCCESS;
    }
    return FAILURE;
}


// WSQ extension function "wsq_convert_image"
// Function reads from one file and writes to another file.

PHP_FUNCTION(wsq_convert_image)
{
    char *input_file_name;
    char *output_file_name;

#if PHP_MAJOR_VERSION >= 7
    size_t length_of_input_filename;
    size_t length_of_output_filename;
#else //if PHP_MAJOR_VERSION >= 7
    int length_of_input_filename;
    int length_of_output_filename;
#endif //#if PHP_MAJOR_VERSION >= 7

    char *input_path_copy = NULL;
    char *output_path_copy = NULL;

    int width = 0;
    int height = 0;
    unsigned char* imageData = 0;
    long type;


    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ssl", &input_file_name, &length_of_input_filename,
				&output_file_name, &length_of_output_filename, &type) == FAILURE) {
        RETURN_FALSE;
    }

    // Function accepts only formats from 1 to 7
    if ( ( type < 1 ) || ( type > 7 ) ) {
	php_printf("Error: output format type must be a number beetween 1 and 7. But it is %ld\n",type);
	RETURN_FALSE;
    }

// Lets check if we could read/write from/to files
#ifdef VIRTUAL_DIR
    virtual_filepath_ex(input_file_name, &input_path_copy, NULL TSRMLS_CC);
    virtual_filepath_ex(output_file_name, &output_path_copy, NULL TSRMLS_CC);
#else
    input_path_copy = input_file_name;
    output_path_copy = output_file_name;
#endif  

    if (php_check_open_basedir(input_path_copy TSRMLS_CC)) {
        RETURN_FALSE;
    }

    if (php_check_open_basedir(output_path_copy TSRMLS_CC)) {
        RETURN_FALSE;
    }

    // Now call WSQ library function and read image from file
    if ( !ReadImageFromFile(input_file_name, &width, &height, &imageData) ) {
	php_printf("Error: WSQ ReadImageFromFile failed!\n");
	RETURN_FALSE;
    }
    


//  type = 1; //WSQ;
//  type = 2; //BMP;
//  type = 3; //TIF;
//  type = 4; //PNG;
//  type = 5; //JPG;
//  type = 6; //RGB;
//  type = 7; //TGA;

    
    // Now call WSQ library function and write image array to file
    if ( !SaveImageToFile(output_file_name, (int) type, width, height, imageData) ) {
	php_printf("Error: WSQ SaveImageToFile failed!\n");
	RETURN_FALSE;
    }

    RETURN_TRUE;
}




// WSQ extension function "wsq_read_image_from_file"
// Function reads from one file to special array.
//
// Parameters are: string fileName

#if PHP_MAJOR_VERSION >= 7
PHP_FUNCTION(wsq_read_image_from_file)
{
    char *input_file_name;
    size_t length_of_input_filename;
    char* input_path_copy = NULL;
    
    int width = 0;
    int height = 0;
    unsigned char* imageData = 0;
    int type;

    long arraySize;
    long i;

    int pixel;
    
    // Subarrays
    zval data, info;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &input_file_name, &length_of_input_filename) == FAILURE) {
        RETURN_FALSE;
    }

// Lets check if we could read from file
#ifdef VIRTUAL_DIR
    virtual_filepath_ex(input_file_name, &input_path_copy, NULL TSRMLS_CC);
#else
    input_path_copy = input_file_name;
#endif  

    if (php_check_open_basedir(input_path_copy TSRMLS_CC)) {
        RETURN_FALSE;
    }
    
    // Now call WSQ library function and read image from file
    if ( !ReadImageFromFile(input_file_name, &width, &height, &imageData) ) {
	php_printf("Error: WSQ ReadImageFromFile failed!\n");
	RETURN_FALSE;
    }
    
    // It's hidden value, It's used to return value to PHP
    array_init(return_value);
    
    arraySize = width*height;
    
    // Alocating 2 subarrays
    array_init(&info);
    array_init(&data);
    
    // Lets fill in info array
    add_assoc_long(&info,"width",width);
    add_assoc_long(&info,"height",height);
    // Lets associate return array with info
    add_assoc_zval(return_value, "info", &info);
    
    // Now we will fill up data array
    for(i = 0; i < arraySize; i++) {
	pixel = ((int)imageData[i] << 16) | ((int)imageData[i+arraySize] << 8) | ((int)imageData[i+2*arraySize]) ;
	add_next_index_long(&data, (long) pixel );		
    }
    // Lets associate return array with data
    add_assoc_zval(return_value, "data", &data);
}


#else //if PHP_MAJOR_VERSION >= 7


PHP_FUNCTION(wsq_read_image_from_file)
{
    char *input_file_name;
    int length_of_input_filename;
    char* input_path_copy = NULL;
    
    int width = 0;
    int height = 0;
    unsigned char* imageData = 0;
    int type;

    long arraySize;
    long i;

    int pixel;
    
    // Subarrays
    zval *data = NULL, *info = NULL;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &input_file_name, &length_of_input_filename) == FAILURE) {
        RETURN_FALSE;
    }

// Lets check if we could read from file
#ifdef VIRTUAL_DIR
    virtual_filepath_ex(input_file_name, &input_path_copy, NULL TSRMLS_CC);
#else
    input_path_copy = input_file_name;
#endif  

    if (php_check_open_basedir(input_path_copy TSRMLS_CC)) {
        RETURN_FALSE;
    }
    
    // Now call WSQ library function and read image from file
    if ( !ReadImageFromFile(input_file_name, &width, &height, &imageData) ) {
	php_printf("Error: WSQ ReadImageFromFile failed!\n");
	RETURN_FALSE;
    }
    
    // It's hidden value, It's used to return value to PHP
    array_init(return_value);
    
    arraySize = width*height;
    
    // Alocating 2 subarrays
    ALLOC_INIT_ZVAL(info);
    array_init(info);
    ALLOC_INIT_ZVAL(data);
    array_init(data);
    
    // Lets fill in info array
    add_assoc_long(info,"width",width);
    add_assoc_long(info,"height",height);
    // Lets associate return array with info
    add_assoc_zval(return_value, "info", info);
    
    // Now we will fill up data array
    for(i=0; i < arraySize; i++ ) {
	pixel = ((int)imageData[i] << 16) | ((int)imageData[i+arraySize] << 8) | ((int)imageData[i+2*arraySize]) ;
	add_next_index_long(data, (long) pixel );		
    }
    // Lets associate return array with data
    add_assoc_zval(return_value, "data", data);
}


#endif //#if PHP_MAJOR_VERSION >= 7



long zval_to_long(zval *value)
{
	if (Z_TYPE_P(value) != IS_LONG) {
		zval tmp_member;
		tmp_member = *value;
		zval_copy_ctor(&tmp_member);
		convert_to_long(&tmp_member);
		long val = Z_LVAL(tmp_member);
		zval_dtor(&tmp_member);
		return val;
	} else {
		return Z_LVAL(*value);
	}
}



// WSQ extension function "wsq_save_image_to_file"
// Functions reads from one file to special array.
//
// Parameters are: array, outputName, outputType
// 2 first elements in the array are width and height


#if PHP_MAJOR_VERSION >= 7

PHP_FUNCTION(wsq_save_image_to_file)
{
    zval *arr, *data;

    HashTable *ht;
    zval *val;
    void *ptr;

    int array_count = 0;

    unsigned char* imageData = NULL;
    long number_of_pixels = 0;
    long width = 0, height = 0, type = 0;

    char *output_file_name;
    size_t length_of_output_filename;
    char* output_path_copy = NULL;
    
    long i = 0, i2 = 0, i3 = 0, error = 0;


    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "asl", &arr, 
	    &output_file_name, &length_of_output_filename, &type) == FAILURE) {
        RETURN_FALSE;
    }


    if ( ( type < 1 ) || ( type > 7 ) ) {
	php_printf("Error: output format type must be a number beetween 1 and 7. But it is %ld\n",type);
	RETURN_FALSE;
    }

// Lets check if we could write to file
#ifdef VIRTUAL_DIR
    virtual_filepath_ex(output_file_name, &output_path_copy, NULL TSRMLS_CC);
#else
    output_path_copy = output_file_name;
#endif  

    if (php_check_open_basedir(output_path_copy TSRMLS_CC)) {
        RETURN_FALSE;
    }
    
    ht = Z_ARRVAL_P(arr);
    array_count = zend_hash_num_elements(ht);
    
    if ( array_count != 2 ) {
	php_printf("Error: invalid wsq image array, input was recieved as %ld items\n",array_count);
	RETURN_FALSE;
    }



//https://wiki.php.net/phpng-upgrading
//ZEND_HASH_FOREACH_VAL(ht, val)
//ZEND_HASH_FOREACH_KEY(ht, h, key)
//ZEND_HASH_FOREACH_PTR(ht, ptr)
//ZEND_HASH_FOREACH_NUM_KEY(ht, h)
//ZEND_HASH_FOREACH_STR_KEY(ht, key)
//ZEND_HASH_FOREACH_STR_KEY_VAL(ht, key, val)
//ZEND_HASH_FOREACH_KEY_VAL(ht, h, key, val)

i = 0;
i2 = 0;
i3 = 0;

ZEND_HASH_FOREACH_PTR(ht, ptr) {
 if(i == 0)
 {
  ZEND_HASH_FOREACH_VAL((HashTable *)ptr, val) {
  
  if (Z_TYPE_P(val) != IS_LONG)
  {
   php_printf("Error: image width and height are invalid, not an integer.\n");
   RETURN_FALSE;
  }

  // php_debug_zval_dump(val, 0);
  // php_printf("val = %ld\n", zval_to_long(val));
  if(i2 == 0) 
  {  
    width = zval_to_long(val);
  }
  else if(i2 == 1)
  {
    height = zval_to_long(val);

    number_of_pixels = width*height;

    imageData = (unsigned char*) emalloc(sizeof(unsigned char)*number_of_pixels*3);
    if (imageData == NULL) {
	php_printf("Error: insuficient memory for imageData.\n");
	return;
    }
  };//else if(i2 == 1)

  i2++;
 } ZEND_HASH_FOREACH_END();
}
else if(i == 1)
{
 ZEND_HASH_FOREACH_VAL((HashTable *)ptr, val) {
//     php_debug_zval_dump(val, 0);
        if (Z_TYPE_P(val) == IS_LONG) {
	    // Here we fill 1 pixel with RGB data from file
	    *(imageData+i3) = (unsigned char)((((int)Z_LVAL_P(val))>>16) & 0x000FF);
    	    *(imageData+i3+number_of_pixels) = (unsigned char)((((int)Z_LVAL_P(val))>>8) & 0x000FF);
    	    *(imageData+i3+2*number_of_pixels) = (unsigned char)((int)Z_LVAL_P(val) & 0x000FF);
        } else {
	    php_printf("Error: invalid wsq image array\n");
	    efree(imageData);
	    return;
	}
 i3++;
 } ZEND_HASH_FOREACH_END();
};//else if(i == 1)


i++;
} ZEND_HASH_FOREACH_END();




    if ( error ) {
	php_printf("Error: invalid wsq image array\n");
	RETURN_FALSE;
    }    
    

//    php_printf("Output arguments of function wsq_save_image_to_file is: width=%ld, height=%ld\n",width,height);


    // Now call WSQ library function and write image array to file
    if ( !SaveImageToFile(output_file_name, (int) type, (int) width, (int) height, imageData) ) {
	php_printf("Error: WSQ SaveImageToFile failed!\n");
	efree(imageData);
	RETURN_FALSE;
    }
    
    efree(imageData);
    RETURN_TRUE;
}



#else //if PHP_MAJOR_VERSION >= 7


PHP_FUNCTION(wsq_save_image_to_file)
{
    zval *arr, **data;
    HashTable *arr_hash;
    HashPosition pointer;
    int array_count;

    zval **data2;
    HashTable *arr_hash2;
    HashPosition pointer2;

    zval **data3;
    HashTable *arr_hash3;
    HashPosition pointer3;
    int array_count3;

    char *output_file_name;
    int length_of_output_filename;
    char* output_path_copy = NULL;
    
    long width, height, type;
    unsigned char* imageData = NULL;

    long number_of_pixels;
    long i = 0, i2 = 0, i3 = 0, error = 0;


    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "asl", &arr, 
	    &output_file_name, &length_of_output_filename, &type) == FAILURE) {
        RETURN_FALSE;
    }

    if ( ( type < 1 ) || ( type > 7 ) ) {
	php_printf("Error: output format type must be a number beetween 1 and 7. But it is %ld\n",type);
	RETURN_FALSE;
    }

// Lets check if we could write to file
#ifdef VIRTUAL_DIR
    virtual_filepath_ex(output_file_name, &output_path_copy, NULL TSRMLS_CC);
#else
    output_path_copy = output_file_name;
#endif  

    if (php_check_open_basedir(output_path_copy TSRMLS_CC)) {
        RETURN_FALSE;
    }
    
    arr_hash = Z_ARRVAL_P(arr);
    array_count = zend_hash_num_elements(arr_hash);
    
    if ( array_count != 2 ) {
	php_printf("Error: invalid wsq image array, input was recieved as %ld items\n", (long)array_count);
	RETURN_FALSE;
    }

for(zend_hash_internal_pointer_reset_ex(arr_hash, &pointer), i=0; 
   zend_hash_get_current_data_ex(arr_hash, (void**) &data, &pointer) == SUCCESS; 
   zend_hash_move_forward_ex(arr_hash, &pointer),i++) {

  if (i == 0)
  {
//===============================
if (Z_TYPE_P(*data) != IS_ARRAY) 
{
php_printf("Error: Data is not an array.\n");
RETURN_FALSE;
}

    arr_hash2 = Z_ARRVAL_P(*data);
    
for(zend_hash_internal_pointer_reset_ex(arr_hash2, &pointer2), i2=0;
   zend_hash_get_current_data_ex(arr_hash2, (void**) &data2, &pointer2) == SUCCESS; 
   zend_hash_move_forward_ex(arr_hash2, &pointer2),i2++) {

  if (i2 == 0)
  {

   if (Z_TYPE_PP(data2) != IS_LONG)
   {
    php_printf("Error: image width is invalid, not an integer.\n");
    RETURN_FALSE;
   }

   width = Z_LVAL_PP(data2);

  }
  else if (i2 == 1)
  {
   if (Z_TYPE_PP(data2) != IS_LONG)
   {
    php_printf("Error: image height is invalid, not an integer.\n");
    RETURN_FALSE;
   }

   height = Z_LVAL_PP(data2);
  }

}//for(zend_hash_internal_pointer_reset_ex(arr_hash2, &pointer2), i2=0;
//===============================
   number_of_pixels = width*height;

  }
  else if (i == 1)
  {
//==============================
    if (Z_TYPE_P(*data) != IS_ARRAY) 
   {
     php_printf("Error: Data is not an array.\n");
     RETURN_FALSE;
    }
    
    arr_hash3 = Z_ARRVAL_P(*data);
    array_count3 = zend_hash_num_elements(arr_hash3);
    
    if (array_count3 != number_of_pixels) 
    {
	php_printf("Error: array elements count (%ld) does not match width*height (%ld).\n", (long)array_count3, number_of_pixels);
	RETURN_FALSE;
    }
    
    imageData = (unsigned char*) emalloc(sizeof(unsigned char)*number_of_pixels*3);
    if (imageData == NULL) {
	php_printf("Error: insuficient memory for imageData.\n");
	RETURN_FALSE;
    }

    for(zend_hash_internal_pointer_reset_ex(arr_hash3, &pointer3),i3=0;
	zend_hash_get_current_data_ex(arr_hash3, (void**) &data3, &pointer3) == SUCCESS; 
	zend_hash_move_forward_ex(arr_hash3, &pointer3),i3++) {
        if (Z_TYPE_PP(data3) == IS_LONG) 
        {
	    // Here we fill 1 pixel with RGB data3 from file
	    *(imageData+i3) = (unsigned char)((((int)Z_LVAL_PP(data3))>>16) & 0x000FF);
    	    *(imageData+i3+number_of_pixels) = (unsigned char)((((int)Z_LVAL_PP(data3))>>8) & 0x000FF);
    	    *(imageData+i3+2*number_of_pixels) = (unsigned char)((int)Z_LVAL_PP(data3) & 0x000FF);
        }
        else
        {
	    php_printf("Error: invalid wsq image array\n");
	    efree(imageData);
	    RETURN_FALSE;
	}
    }//for(zend_hash_internal_pointer_reset_ex(arr_hash3, &pointer3),i3=0;

//==============================


    if ( imageData == NULL ) 
    {
	error = 1;
	break;
    }


  }//else if (i == 1)

}//for(zend_hash_internal_pointer_reset_ex(arr_hash, &pointer), i=0;

    if ( error ) {
	php_printf("Error: invalid wsq image array\n");
	RETURN_FALSE;
    }    
    
    // Now call WSQ library function and write image array to file
    if ( !SaveImageToFile(output_file_name, (int) type, (int) width, (int) height, imageData) ) {
	php_printf("Error: WSQ SaveImageToFile failed!\n");
	efree(imageData);
	RETURN_FALSE;
    }
    
    efree(imageData);
    RETURN_TRUE;
}


#endif //#if PHP_MAJOR_VERSION >= 7




// WSQ extension function "wsq_write_wsq_bitrate"
//
// Parameters are: double bitrate

PHP_FUNCTION(wsq_write_wsq_bitrate)
{
    zval *arr;

    HashTable *arr_hash;
    HashPosition pointer;
    int array_count;

    double bitrate = 0.75;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "d",
	    &bitrate) == FAILURE) {
        RETURN_FALSE;
    }

    // Now call WSQ library function and decode stream
    WriteWSQ_bitrate(bitrate);

    RETURN_TRUE;

};


// WSQ extension function "wsq_read_wsq_bitrate"

PHP_FUNCTION(wsq_read_wsq_bitrate)
{

    zval *arr;

    HashTable *arr_hash;
    HashPosition pointer;
    int array_count;

    double bitrate = 0.75;


    // Now call WSQ library function and decode stream
    if ( ! (bitrate = ReadWSQ_bitrate()) ) {
	php_printf("Error: WSQ ReadWSQ_bitrate failed!\n");
	RETURN_FALSE;
    }

    RETURN_DOUBLE(bitrate);

};


// WSQ extension function "wsq_write_wsq_ppi"
//
// Parameters are: int ppi


PHP_FUNCTION(wsq_write_wsq_ppi)
{

    zval *arr;

    HashTable *arr_hash;
    HashPosition pointer;
    int array_count;

    long ppi = -1;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l",
	    &ppi) == FAILURE) {
        RETURN_FALSE;
    }

     WriteWSQ_ppi((int)ppi);

    RETURN_TRUE;

};

//   WSQ extension function "wsq_read_wsq_ppi"


PHP_FUNCTION(wsq_read_wsq_ppi)
{

    zval *arr;

    HashTable *arr_hash;
    HashPosition pointer;
    int array_count;

    long ppi = -1;


    if ( ! (ppi = (long)ReadWSQ_ppi()) ) {
	php_printf("Error: WSQ ReadWSQ_ppi failed!\n");
	RETURN_LONG(-1);
    }

    RETURN_LONG(ppi);

};


//   WSQ extension function "wsq_write_wsq_comment"
//
//   Parameters are: string wsq_comment

PHP_FUNCTION(wsq_write_wsq_comment)
{

    zval *arr;

    HashTable *arr_hash;
    HashPosition pointer;
    int array_count;

    char* wsq_comment;
    size_t length_of_wsq_comment;


    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s",
	    &wsq_comment, &length_of_wsq_comment) == FAILURE) {
        RETURN_FALSE;
    }

     WriteWSQ_comment(wsq_comment);

    RETURN_TRUE;

};


//   WSQ extension function "wsq_read_wsq_comment"


PHP_FUNCTION(wsq_read_wsq_comment)
{

    zval *arr;

    HashTable *arr_hash;
    HashPosition pointer;
    int array_count;

    char* wsq_comment;

    if ( !(wsq_comment = ReadWSQ_comment()) ) {
	php_printf("Error: WSQ ReadWSQ_comment failed!\n");
	RETURN_FALSE;
    }

#if PHP_MAJOR_VERSION >= 7
   RETURN_STRING(wsq_comment);
#else //if PHP_MAJOR_VERSION >= 7
   RETURN_STRING(wsq_comment, 1);
#endif //#if PHP_MAJOR_VERSION >= 7


};


// WSQ extension function "wsq_write_tiff_compression"
//
// Parameters are: int tiff_compression

PHP_FUNCTION(wsq_write_tiff_compression)
{
    zval *arr;

    HashTable *arr_hash;
    HashPosition pointer;
    int array_count;

    long tiff_compression = 1;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l",
	    &tiff_compression) == FAILURE) {
        RETURN_FALSE;
    }

     WriteTIFFcompression((int)tiff_compression);

    RETURN_TRUE;
};


// WSQ extension function "wsq_write_tiff_predictor"
//
// Parameters are int tiff_predictor


PHP_FUNCTION(wsq_write_tiff_predictor)
{
    zval *arr;

    HashTable *arr_hash;
    HashPosition pointer;
    int array_count;

    long tiff_predictor = 1;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l",
	    &tiff_predictor) == FAILURE) {
        RETURN_FALSE;
    }

     WriteTIFFpredictor((int)tiff_predictor);

    RETURN_TRUE;
};


//   WSQ extension function "wsq_generate_serial_number"

PHP_FUNCTION(wsq_generate_serial_number)
{
    zval *arr;

    HashTable *arr_hash;
    HashPosition pointer;
    int array_count;

    char* serial_number;

    if ( !(serial_number = GenerateSerialNumber()) ) {
	php_printf("Error: WSQ GenerateSerialNumber failed!\n");
	RETURN_FALSE;
    }

#if PHP_MAJOR_VERSION >= 7
    RETURN_STRING(serial_number);
#else //if PHP_MAJOR_VERSION >= 7
    RETURN_STRING(serial_number, 1);
#endif //#if PHP_MAJOR_VERSION >= 7

};


//   WSQ extension function "wsq_unlock_wsq_library"
//
//   Parameters are: string unlock_code


PHP_FUNCTION(wsq_unlock_wsq_library)
{

    zval *arr;

    HashTable *arr_hash;
    HashPosition pointer;
    int array_count;
    int ret;

    char* unlock_code;
    size_t len_unlock_code;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s",
	    &unlock_code, &len_unlock_code) == FAILURE) {
        RETURN_FALSE;
    }

    ret = UnlockWSQLibrary(unlock_code);

    RETURN_LONG((long)ret);
};





#if PHP_MAJOR_VERSION >= 7


// WSQ extension function "wsq_decode_stream"
// Reads from stream and returns decoded stream.
//
// Parameters are stream. It will return array. Values in array
// will be: array "info" where you will fin "width", "height", "ppi", "Comment"
// Binary String of data (to save space will not return array of long)


PHP_FUNCTION(wsq_decode_stream)
{
    zval *zstream = NULL, *data = NULL;
    zval info;
    php_stream *stream;
    zend_string *zdata_in = NULL;
    char *data_in = NULL;
    unsigned char *data_out = NULL;
    int width = 0, height = 0, ppi = 0, size = 0, i = 0;
    char* comment = NULL; //prevents crashing

        
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zstream) == FAILURE ) {
	RETURN_FALSE;
    }


    php_stream_from_zval(stream, zstream);


    if ( !stream ) {
	php_printf("Error: invalid stream.");
	RETURN_FALSE;
    }

    // Lets copy from stream to memory buffer
    zdata_in = php_stream_copy_to_mem(stream,PHP_STREAM_COPY_ALL,0);
    data_in = ZSTR_VAL(zdata_in);
    size = ZSTR_LEN(zdata_in);

if ( !data_in || ( size == 0 ) ) {
	php_printf("Error: failed ZSTR_VAL.");
	RETURN_FALSE;
    }
//php_printf("size =  %d\n", size);



    // Now call WSQ library function and decode stream
    if ( !WSQ_decode_stream((unsigned char*) data_in, size, &data_out, &width, &height, &ppi, &comment) ) {
	php_printf("Error: WSQ WSQ_decode_stream failed!\n");
	//pefree(data_in,0); // don't do pefree in PHP 7 
	RETURN_FALSE;
    }

    //php_printf("Output of function wsq_decode_stream is: width=%ld, height=%ld, ppi=%ld, comment=%s\n",width,height,ppi,comment);

    // Lets release data 
    //pefree(data_in,0); // don't do pefree in PHP 7 


    if ( !data_out || (width*height == 0) ) {
	php_printf("Error: WSQ WSQ_decode_stream failed!\n");
	RETURN_FALSE;
    }
    // Lets return decoded stream to PHP 
    
    // It's hidden value, It's used to return value to PHP 
    array_init(return_value);
    array_init(&info);


    // Lets fill in info into array
    add_assoc_long(&info,"width",width);
    add_assoc_long(&info,"height",height);
    add_assoc_long(&info,"ppi",ppi);
    if ( comment ) {
	add_assoc_string(&info, "comment", comment);
	free(comment);
    }


    // Lets associate return array with info 
    add_assoc_zval(return_value,"info",&info);
    // Now we create return array for data 
    add_assoc_stringl(return_value, "data", data_out, width*height);

    free(data_out);
}



// WSQ extension function "wsq_encode_stream"
// It will read from stream and will return encoded stream.
//
// Parameters are stream

PHP_FUNCTION(wsq_encode_stream)
{
    zval *zstream = NULL, *arr = NULL, *data, *data2;
    php_stream *stream;
    unsigned char *imageData  = NULL, *data_out = NULL;
    long width = 0, height = 0, ppi = 0;
    int size = 0;
    char* comment = NULL;
    double bitrate = 0.0;


    HashTable *ht, *ht2;
    zval *val, *val2;

    HashPosition pointer, pointer2;


    long i = 0, i2 = 0, i3 = 0;


    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rad", &zstream, &arr, &bitrate) == FAILURE ) {
	RETURN_FALSE;
    }
    
    php_stream_from_zval(stream, zstream);

    if ( !stream ) {
	php_printf("Error: invalid stream.");
	RETURN_FALSE;
    }


    ht = Z_ARRVAL_P(arr);
    size = zend_hash_num_elements(ht);

  
    if ( size != 2 ) {
	php_printf("Error: invalid wsq stream encoding array, input was recieved as %ld items\n", (long)size);
	RETURN_FALSE;
    }


//https://wiki.php.net/phpng-upgrading
//ZEND_HASH_FOREACH_VAL(ht, val)
//ZEND_HASH_FOREACH_KEY(ht, h, key)
//ZEND_HASH_FOREACH_PTR(ht, ptr)
//ZEND_HASH_FOREACH_NUM_KEY(ht, h)
//ZEND_HASH_FOREACH_STR_KEY(ht, key)
//ZEND_HASH_FOREACH_STR_KEY_VAL(ht, key, val)
//ZEND_HASH_FOREACH_KEY_VAL(ht, h, key, val)



//One way to parse arrays is below:
//========================================
i = 0;
i2 = 0;
i3 = 0;

ZEND_HASH_FOREACH_VAL(ht, val) {
 if(i == 0)
 {

  ht2 = Z_ARRVAL_P(val);

  ZEND_HASH_FOREACH_VAL(ht2, val2) {
  
  // php_debug_zval_dump(val2, 0);
  // php_printf("val2 = %ld\n", zval_to_long(val2));
  if(i2 == 0) 
  {  
     if (Z_TYPE_P(val2) != IS_LONG)
     {
      php_printf("Error: image width is invalid, not an integer.\n");
      RETURN_FALSE;
     }

     width = zval_to_long(val2);
  }

  else if(i2 == 1)
  {
     if (Z_TYPE_P(val2) != IS_LONG)
     {
      php_printf("Error: image height is invalid, not an integer.\n");
      RETURN_FALSE;
     }

     height = zval_to_long(val2);  

  }//else if(i2 == 1)

  else if(i2 == 2)
  {
     if (Z_TYPE_P(val2) != IS_LONG)
     {
      php_printf("Error: image ppi is invalid, not an integer.\n");
      RETURN_FALSE;
     }

     ppi = zval_to_long(val2);
  }//else if(i2 == 2)
  else if(i2 == 3)
  {
     if (Z_TYPE_P(val2) != IS_STRING)
     {
      php_printf("Error: image comment is invalid, not a zend_string.\n");
      RETURN_FALSE;
     }

     //It is read not binary safe - but WSQ_encode_stream doesn't expect it to be binary string 	 
     comment = Z_STRVAL_P(val2); 
  }//else if(i2 == 3)



  i2++;
 } ZEND_HASH_FOREACH_END();


}
else if(i == 1)
{
//php_debug_zval_dump(val, 0);

     if (Z_TYPE_P(val) != IS_STRING)
     {
      php_printf("Error: imageData is invalid, not a zend_string.\n");
      RETURN_FALSE;
     }

     imageData = Z_STRVAL_P(val);
     size      = Z_STRLEN_P(val);

};//else if(i == 1)



i++;
} ZEND_HASH_FOREACH_END();
//========================================



/*
//Alternative way to parse arrays is below:
//========================================
i = 0;
i2 = 0;
i3 = 0;


for(zend_hash_internal_pointer_reset_ex(ht, &pointer), i=0;
   (data = zend_hash_get_current_data_ex(ht, &pointer)) != NULL; 
   zend_hash_move_forward_ex(ht, &pointer),i++) 
{
// Now lets read 2 arrays from our array
  if (i == 0) 
  {
//  php_debug_zval_dump(data, 0);
    if (Z_TYPE_P(data) != IS_ARRAY) {
      php_printf("Error: data is not an array.\n");
      RETURN_FALSE;
    }

   ht2 = Z_ARRVAL_P(data);

   for(zend_hash_internal_pointer_reset_ex(ht2, &pointer2),i2=0;
   (data2 = zend_hash_get_current_data_ex(ht2, &pointer2)) != NULL; 
   zend_hash_move_forward_ex(ht2, &pointer2),i2++) {	

    if(i2 == 0) 
    {  
     if (Z_TYPE_P(data2) != IS_LONG)
     {
      php_printf("Error: image width is invalid, not an integer.\n");
      RETURN_FALSE;
     }

     width = zval_to_long(data2);
    }

    else if(i2 == 1)
    {
     if (Z_TYPE_P(data2) != IS_LONG)
     {
      php_printf("Error: image height is invalid, not an integer.\n");
      RETURN_FALSE;
     }

     height = zval_to_long(data2);   

    }//else if(i2 == 1)

    else if(i2 == 2)
    {
     if (Z_TYPE_P(data2) != IS_LONG)
     {
      php_printf("Error: image ppi is invalid, not an integer.\n");
      RETURN_FALSE;
     }

     ppi = zval_to_long(data2);
    }//else if(i2 == 2)
    else if(i2 == 3)
    {
     if (Z_TYPE_P(data2) != IS_STRING)
     {
      php_printf("Error: image comment is invalid, not a zend_string.\n");
      RETURN_FALSE;
     }

     //It is read not binary safe - but WSQ_encode_stream doesn't expect it to be binary string 	 
     comment = Z_STRVAL_P(data2); 
    }//else if(i2 == 3)

   }//for(zend_hash_internal_pointer_reset_ex(ht2, &pointer2),i2=0;


  }//if (i == 0) 
  else if (i== 1) 
  {
     if (Z_TYPE_P(data) != IS_STRING)
     {
      php_printf("Error: imageData is invalid, not a zend_string.\n");
      RETURN_FALSE;
     }

     imageData = Z_STRVAL_P(data);
     size      = Z_STRLEN_P(data);
  }//else if (i== 1) 
}//for(zend_hash_internal_pointer_reset_ex(ht, &pointer), i=0;
//========================================
*/



//php_printf("Output of function wsq_encode_stream is: width=%ld, height=%ld, ppi=%ld, comment=%s, size=%ld\n", width, height, ppi, comment, size);



if ( !imageData || !size ) 
{
 php_printf("Error: invalid wsq image array\n");
 RETURN_FALSE;
}


    
    if (bitrate == 0.0)
	bitrate = 0.75;
    
    // Now we will encode data
    if ( WSQ_encode_stream(imageData, (int) width, (int) height, bitrate, (int) ppi, comment, &data_out, &size) ) 
    {
	php_printf("Error: WSQ WSQ_encode_stream failed! Data pointer is %ld, size is %ld!\n", (long)*data_out, (long)size);
	RETURN_FALSE;
    }

    if (!data_out || !size) 
    {
	php_printf("Error: WSQ WSQ_encode_stream failed, data pointer is %ld, size is %ld!\n", (long)*data_out, (long)size);
	RETURN_FALSE;
    }
    // now we will write to stream
    width = php_stream_write(stream,data_out,size);
    if (width != size) 
    {
	php_printf("Error: writing to stream failed, wrote only %ld from %ld!\n", width, size);
	RETURN_FALSE;
    }
    free(data_out);

    
    RETURN_TRUE;
}


#else //if PHP_MAJOR_VERSION >= 7


// WSQ extension function "wsq_decode_stream"
// Reads from stream and returns decoded stream.
//
// Parameters are stream. It will return array. Values in array
// will be: array "info" where you will fin "width", "height", "ppi", "Comment"
// Binary String of data (to save space will not return array of long)


PHP_FUNCTION(wsq_decode_stream)
{
    zval *zstream = NULL, *info = NULL, *data = NULL;
    php_stream *stream;
    char *data_in;
    unsigned char *data_out;
    int width = 0, height = 0, ppi = 0, size = 0, i = 0;
    char* comment = NULL; //prevents crashing
        
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zstream) == FAILURE ) {
	RETURN_FALSE;
    }
    
    php_stream_from_zval(stream, &zstream);

    if ( !stream ) {
	php_printf("Error: invalid stream.");
	RETURN_FALSE;
    }
    // Lets copy from stream to memory buffer
    size = php_stream_copy_to_mem(stream, &data_in, PHP_STREAM_COPY_ALL, 0);
    if ( !data_in || ( size == 0 ) ) {
	php_printf("Error: unable to allocate memmory for stream data.");
	RETURN_FALSE;
    }
    
    // Now call WSQ library function and decode stream
    if ( !WSQ_decode_stream((unsigned char*) data_in, size, &data_out, &width, &height, &ppi, &comment) ) {
	php_printf("Error: WSQ WSQ_decode_stream failed!\n");
	pefree(data_in,0);
	RETURN_FALSE;
    }
//    php_printf("Input of function wsq_decode_stream is: width=%ld, height=%ld, ppi=%ld, comment=%s\n", width, height, ppi, comment);
    // Lets release data
    pefree(data_in,0);

    if ( !data_out || (width*height == 0) ) 
    {
	php_printf("Error: WSQ WSQ_decode_stream failed!\n");
	RETURN_FALSE;
    }
    // Lets return decoded stream to PHP
    
    // It's hidden value, It's used to return value to PHP
    array_init(return_value);
    ALLOC_INIT_ZVAL(info);
    array_init(info);
    
    // Lets fill in info into array
    add_assoc_long(info, "width", width);
    add_assoc_long(info, "height", height);
    add_assoc_long(info, "ppi", ppi);
    if ( comment ) 
    {
	add_assoc_string(info, "comment", comment, 1);
	free(comment);
    }
    // Lets associate return array with info
    add_assoc_zval(return_value, "info", info);
    // Now we create return array for data
    add_assoc_stringl(return_value, "data", data_out, width*height, 1);
    free(data_out);
}


// WSQ extension function "wsq_encode_stream"
// It will read from stream and will return encoded stream.
//
// Parameters are stream

PHP_FUNCTION(wsq_encode_stream)
{
    zval *zstream = NULL, *arr = NULL, **data;
    php_stream *stream;
    unsigned char *imageData, *data_out;
    long width = 0, height = 0, ppi = 0;
    int size = 0;
    char* comment = NULL;
    double bitrate;
    
    HashTable *arr_hash;
    HashPosition pointer;
    long error = 0;

    zval **data2;
    HashTable *arr_hash2;
    HashPosition pointer2;
    long i = 0, i2 = 0;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rad", &zstream, &arr,&bitrate) == FAILURE ) 
    {
	RETURN_FALSE;
    }
    
    php_stream_from_zval(stream, &zstream);

    if (!stream) 
    {
	php_printf("Error: invalid stream.");
	RETURN_FALSE;
    }


    arr_hash = Z_ARRVAL_P(arr);
    size = zend_hash_num_elements(arr_hash);
    
    if (size != 2)
    {
	php_printf("Error: invalid wsq stream encoding array, input was recieved as %ld items\n", (long)size);
	RETURN_FALSE;
    }

for(zend_hash_internal_pointer_reset_ex(arr_hash, &pointer), i=0; 
   zend_hash_get_current_data_ex(arr_hash, (void**) &data, &pointer) == SUCCESS; 
   zend_hash_move_forward_ex(arr_hash, &pointer),i++) {
	// Now lets read 2 arrays from our array
	if (i == 0) 
        {
//===============================
if (Z_TYPE_P(*data) != IS_ARRAY) 
{
php_printf("Error: Data is not an array.\n");
RETURN_FALSE;
}

    arr_hash2 = Z_ARRVAL_P(*data);
    
for(zend_hash_internal_pointer_reset_ex(arr_hash2, &pointer2), i2=0;
   zend_hash_get_current_data_ex(arr_hash2, (void**) &data2, &pointer2) == SUCCESS; 
   zend_hash_move_forward_ex(arr_hash2, &pointer2),i2++) {

  if (i2 == 0)
  {

   if (Z_TYPE_PP(data2) != IS_LONG)
   {
    php_printf("Error: image width is invalid, not an integer.\n");
    RETURN_FALSE;
   }

   width = Z_LVAL_PP(data2);

  }
  else if (i2 == 1)
  {
   if (Z_TYPE_PP(data2) != IS_LONG)
   {
    php_printf("Error: image height is invalid, not an integer.\n");
    RETURN_FALSE;
   }

   height = Z_LVAL_PP(data2);
  }
  else if (i2 == 2)
  {
   if (Z_TYPE_PP(data2) != IS_LONG)
   {
    php_printf("Error: image ppi is invalid, not an integer.\n");
    RETURN_FALSE;
   }

   ppi = Z_LVAL_PP(data2);
  }

 else if (i2 == 3)
 {
   if (Z_TYPE_PP(data2) != IS_STRING)
   {
    php_printf("Error: image comment is invalid, not a string.\n");
    RETURN_FALSE;
   }

   comment = Z_STRVAL_PP(data2);
 }//if (i2 == 3)

}//for(zend_hash_internal_pointer_reset_ex(arr_hash2, &pointer2), i2=0;
//===============================

} //if (i == 0)
else if (i == 1)
{
   imageData = Z_STRVAL_PP(data);
   size      = Z_STRLEN_PP(data);

   if ( !imageData || !size ) 
   {
    error = 1;
    break;
   }
} // else if (i == 1)

}//for(zend_hash_internal_pointer_reset_ex(arr_hash, &pointer), i=0; 

    if ( error ) {
	php_printf("Error: invalid wsq image array\n");
	RETURN_FALSE;
    }
    
    if (bitrate == 0.0 )
	bitrate = 0.75;
    
    // Now we will encode data
    if ( WSQ_encode_stream(imageData, (int) width, (int) height, bitrate, (int) ppi, comment, &data_out, &size) ) 
    {
	php_printf("Error: WSQ WSQ_encode_stream failed! Data pointer is %ld, size is %ld!\n", (long)*data_out, (long)size);
	RETURN_FALSE;
    }

    if (!data_out || !size) 
    {
	php_printf("Error: WSQ WSQ_encode_stream failed, data pointer is %ld, size is %ld!\n", (long)*data_out, (long)size);
	RETURN_FALSE;
    }
    // now we will write to stream
    width = php_stream_write(stream,data_out,size);
    if (width != size) 
    {
	php_printf("Error: writing to stream failed, wrote only %ld from %ld!\n", width, (long)size);
	RETURN_FALSE;
    }
    free(data_out);
    
    RETURN_TRUE;
}

#endif //#if PHP_MAJOR_VERSION >= 7



#endif
